package jalview.bin;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import jalview.analysis.AlignmentUtils;
import jalview.bin.argparser.Arg;
import jalview.bin.argparser.ArgParser;
import jalview.bin.argparser.ArgParser.Position;
import jalview.bin.argparser.ArgValue;
import jalview.bin.argparser.ArgValues;
import jalview.bin.argparser.ArgValuesMap;
import jalview.bin.argparser.SubVals;
import jalview.datamodel.AlignmentI;
import jalview.datamodel.SequenceI;
import jalview.datamodel.annotations.AlphaFoldAnnotationRowBuilder;
import jalview.gui.AlignFrame;
import jalview.gui.AlignmentPanel;
import jalview.gui.Desktop;
import jalview.gui.Preferences;
import jalview.gui.StructureChooser;
import jalview.gui.StructureViewer;
import jalview.gui.StructureViewer.ViewerType;
import jalview.io.AppletFormatAdapter;
import jalview.io.BioJsHTMLOutput;
import jalview.io.DataSourceType;
import jalview.io.FileFormat;
import jalview.io.FileFormatException;
import jalview.io.FileFormatI;
import jalview.io.FileLoader;
import jalview.io.HtmlSvgOutput;
import jalview.io.IdentifyFile;
import jalview.structure.StructureImportSettings.TFType;
import jalview.structure.StructureSelectionManager;
import jalview.util.HttpUtils;
import jalview.util.MessageManager;
import jalview.util.Platform;

public class Commands
{
  Desktop desktop;

  private boolean headless;

  private ArgParser argParser;

  private Map<String, AlignFrame> afMap;

  private boolean commandArgsProvided = false;

  private boolean argsWereParsed = false;

  public Commands(ArgParser argparser, boolean headless)
  {
    this(Desktop.instance, argparser, headless);
  }

  public Commands(Desktop d, ArgParser argparser, boolean h)
  {
    argParser = argparser;
    headless = h;
    desktop = d;
    afMap = new HashMap<String, AlignFrame>();
    if (argparser != null)
    {
      processArgs(argparser, headless);
    }
  }

  private boolean processArgs(ArgParser argparser, boolean h)
  {
    argParser = argparser;
    headless = h;
    boolean theseArgsWereParsed = false;

    if (argParser != null && argParser.getLinkedIds() != null)
    {
      for (String id : argParser.getLinkedIds())
      {
        ArgValuesMap avm = argParser.getLinkedArgs(id);
        theseArgsWereParsed = true;
        theseArgsWereParsed &= processLinked(id);
        theseArgsWereParsed &= processImages(id);

        // close ap
        if (avm.getBoolean(Arg.CLOSE))
        {
          AlignFrame af = afMap.get(id);
          if (af != null)
          {
            af.closeMenuItem_actionPerformed(true);
          }
        }

      }

    }
    if (argParser.getBool(Arg.QUIT))
    {
      Jalview.getInstance().quit();
      return true;
    }
    // carry on with jalview.bin.Jalview
    argsWereParsed = theseArgsWereParsed;
    return argsWereParsed;
  }

  public boolean commandArgsProvided()
  {
    return commandArgsProvided;
  }

  public boolean argsWereParsed()
  {
    return argsWereParsed;
  }

  protected boolean processUnlinked(String id)
  {
    return processLinked(id);
  }

  protected boolean processLinked(String id)
  {
    boolean theseArgsWereParsed = false;
    ArgValuesMap avm = argParser.getLinkedArgs(id);
    if (avm == null)
      return true;

    /*
     * // script to execute after all loading is completed one way or another String
     * groovyscript = m.get(Arg.GROOVY) == null ? null :
     * m.get(Arg.GROOVY).getValue(); String file = m.get(Arg.OPEN) == null ? null :
     * m.get(Arg.OPEN).getValue(); String data = null; FileFormatI format = null;
     * DataSourceType protocol = null;
     */
    if (avm.containsArg(Arg.APPEND) || avm.containsArg(Arg.OPEN))
    {
      commandArgsProvided = true;
      long progress = -1;

      boolean first = true;
      boolean progressBarSet = false;
      AlignFrame af;
      // Combine the APPEND and OPEN files into one list, along with whether it
      // was APPEND or OPEN
      List<ArgValue> openAvList = new ArrayList<>();
      openAvList.addAll(avm.getArgValueList(Arg.OPEN));
      openAvList.addAll(avm.getArgValueList(Arg.APPEND));
      // sort avlist based on av.getArgIndex()
      Collections.sort(openAvList);
      for (ArgValue av : openAvList)
      {
        Arg a = av.getArg();
        SubVals sv = av.getSubVals();
        String openFile = av.getValue();
        if (openFile == null)
          continue;

        theseArgsWereParsed = true;
        if (first)
        {
          first = false;
          if (!headless && desktop != null)
          {
            desktop.setProgressBar(
                    MessageManager.getString(
                            "status.processing_commandline_args"),
                    progress = System.currentTimeMillis());
            progressBarSet = true;
          }
        }

        if (!Platform.isJS())
        /**
         * ignore in JavaScript -- can't just file existence - could load it?
         * 
         * @j2sIgnore
         */
        {
          if (!HttpUtils.startsWithHttpOrHttps(openFile))
          {
            if (!(new File(openFile)).exists())
            {
              Console.warn("Can't find file '" + openFile + "'");
            }
          }
        }

        DataSourceType protocol = AppletFormatAdapter
                .checkProtocol(openFile);

        FileFormatI format = null;
        try
        {
          format = new IdentifyFile().identify(openFile, protocol);
        } catch (FileFormatException e1)
        {
          Console.error("Unknown file format for '" + openFile + "'");
        }

        af = afMap.get(id);
        // When to open a new AlignFrame
        if (af == null || "true".equals(av.getSubVal("new"))
                || a == Arg.OPEN || format == FileFormat.Jalview)
        {
          /*
           * this approach isn't working yet // get default annotations before opening
           * AlignFrame if (m.get(Arg.SSANNOTATIONS) != null) {
           * Console.debug("##### SSANNOTATIONS=" + m.get(Arg.SSANNOTATIONS).getBoolean());
           * } if (m.get(Arg.NOTEMPFAC) != null) { Console.debug( "##### NOTEMPFAC=" +
           * m.get(Arg.NOTEMPFAC).getBoolean()); } boolean showSecondaryStructure =
           * (m.get(Arg.SSANNOTATIONS) != null) ? m.get(Arg.SSANNOTATIONS).getBoolean() :
           * false; boolean showTemperatureFactor = (m.get(Arg.NOTEMPFAC) != null) ?
           * !m.get(Arg.NOTEMPFAC).getBoolean() : false; Console.debug("##### tempfac=" +
           * showTemperatureFactor + ", showSS=" + showSecondaryStructure);
           * StructureSelectionManager ssm = StructureSelectionManager
           * .getStructureSelectionManager(Desktop.instance); if (ssm != null) {
           * ssm.setAddTempFacAnnot(showTemperatureFactor);
           * ssm.setProcessSecondaryStructure(showSecondaryStructure); }
           */

          Console.debug(
                  "Opening '" + openFile + "' in new alignment frame");
          FileLoader fileLoader = new FileLoader(!headless);

          af = fileLoader.LoadFileWaitTillLoaded(openFile, protocol,
                  format);
          boolean showAnnotations = ArgParser.getFromSubValArgOrPref(avm,
                  Arg.ANNOTATIONS, av.getSubVals(), null,
                  "SHOW_ANNOTATIONS", true);
          af.setAnnotationsVisibility(showAnnotations, false, true);

          // wrap alignment?
          boolean wrap = ArgParser.getFromSubValArgOrPref(avm, Arg.WRAP, sv,
                  null, "WRAP_ALIGNMENT", false);
          af.getCurrentView().setWrapAlignment(wrap);

          // colour aligment?
          String colour = ArgParser.getFromSubValArgOrPref(avm, av,
                  Arg.COLOUR, sv, null, "DEFAULT_COLOUR_PROT", "");

          if ("" != colour)
          {
            af.changeColour_actionPerformed(colour);
          }

          // change alignment frame title
          String title = ArgParser.getFromSubValArgOrPref(avm, av,
                  Arg.TITLE, sv, null, null, null);
          if (title != null)
            af.setTitle(title);

          // show secondary structure annotations?
          boolean showSSAnnotations = ArgParser.getFromSubValArgOrPref(avm,
                  Arg.SSANNOTATIONS, av.getSubVals(), null,
                  "STRUCT_FROM_PDB", true);
          if (avm.getBoolean(Arg.SSANNOTATIONS))
          {
            af.setAnnotationsVisibility(showSSAnnotations, true, false);
            /*
            AlignmentUtils.showOrHideSequenceAnnotations(
                    af.getCurrentView().getAlignment(),
                    Collections.singleton("Secondary Structure"), null,
                    false, false);
             */
          }

          // show temperature factor annotations?
          if (avm.getBoolean(Arg.NOTEMPFAC))
          {
            // do this better (annotation types?)
            List<String> hideThese = new ArrayList<>();
            hideThese.add("Temperature Factor");
            hideThese.add(AlphaFoldAnnotationRowBuilder.LABEL);
            AlignmentUtils.showOrHideSequenceAnnotations(
                    af.getCurrentView().getAlignment(), hideThese, null,
                    false, false);
          }
          else
          /*
           * comment out hacky approach up to here and add this line: if
           * (showTemperatureFactor)
           */
          {
            /*
            if (avm.containsArg(Arg.TEMPFAC_LABEL))
            {
              AlignmentAnnotation aa = AlignmentUtils
                      .getFirstSequenceAnnotationOfType(
                              af.getCurrentView().getAlignment(),
                              AlignmentAnnotation.LINE_GRAPH);
              String label = avm.getValue(Arg.TEMPFAC_LABEL);
              if (aa != null)
              {
                aa.label = label;
              }
              else
              {
                Console.info(
                        "Could not find annotation to apply tempfac_label '"
                                + label);
              }
            }
            */
          }

          // store the AlignFrame for this id
          afMap.put(id, af);

          // is it its own structure file?
          if (format.isStructureFile())
          {
            StructureSelectionManager ssm = StructureSelectionManager
                    .getStructureSelectionManager(Desktop.instance);
            SequenceI seq = af.alignPanel.getAlignment().getSequenceAt(0);
            ssm.computeMapping(false, new SequenceI[] { seq }, null,
                    openFile, DataSourceType.FILE, null, null, null, false);
          }
        }
        else
        {
          Console.debug(
                  "Opening '" + openFile + "' in existing alignment frame");
          DataSourceType dst = HttpUtils.startsWithHttpOrHttps(openFile)
                  ? DataSourceType.URL
                  : DataSourceType.FILE;
          FileLoader fileLoader = new FileLoader(!headless);
          fileLoader.LoadFile(af.getCurrentView(), openFile, dst, null,
                  false);
        }

        Console.debug("Command " + Arg.APPEND + " executed successfully!");

      }
      if (first) // first=true means nothing opened
      {
        if (headless)
        {
          Jalview.exit("Could not open any files in headless mode", 1);
        }
        else
        {
          Console.warn("No more files to open");
        }
      }
      if (progressBarSet && desktop != null)
        desktop.setProgressBar(null, progress);

    }

    // open the structure (from same PDB file or given PDBfile)
    if (!avm.getBoolean(Arg.NOSTRUCTURE))
    {
      AlignFrame af = afMap.get(id);
      if (avm.containsArg(Arg.STRUCTURE))
      {
        commandArgsProvided = true;
        for (ArgValue av : avm.getArgValueList(Arg.STRUCTURE))
        {
          String val = av.getValue();
          SubVals subVals = av.getSubVals();
          SequenceI seq = getSpecifiedSequence(af, subVals);
          if (seq == null)
          {
            // Could not find sequence from subId, let's assume the first
            // sequence in the alignframe
            AlignmentI al = af.getCurrentView().getAlignment();
            seq = al.getSequenceAt(0);
          }

          if (seq == null)
          {
            Console.warn("Could not find sequence for argument "
                    + Arg.STRUCTURE.argString() + "=" + val);
            // you probably want to continue here, not break
            // break;
            continue;
          }
          File structureFile = null;
          if (subVals.getContent() != null
                  && subVals.getContent().length() != 0)
          {
            structureFile = new File(subVals.getContent());
            Console.debug("Using structure file (from argument) '"
                    + structureFile.getAbsolutePath() + "'");
          }
          // TRY THIS
          /*
           * PDBEntry fileEntry = new AssociatePdbFileWithSeq()
           * .associatePdbWithSeq(selectedPdbFileName, DataSourceType.FILE,
           * selectedSequence, true, Desktop.instance);
           * 
           * sViewer = launchStructureViewer(ssm, new PDBEntry[] { fileEntry }, ap, new
           * SequenceI[] { selectedSequence });
           * 
           */
          /* THIS DOESN'T WORK */
          else if (seq.getAllPDBEntries() != null
                  && seq.getAllPDBEntries().size() > 0)
          {
            structureFile = new File(
                    seq.getAllPDBEntries().elementAt(0).getFile());
            Console.debug("Using structure file (from sequence) '"
                    + structureFile.getAbsolutePath() + "'");
          }

          if (structureFile == null)
          {
            Console.warn("Not provided structure file with '" + val + "'");
            continue;
          }

          if (!structureFile.exists())
          {
            Console.warn("Structure file '"
                    + structureFile.getAbsoluteFile() + "' not found.");
            continue;
          }

          Console.debug("Using structure file "
                  + structureFile.getAbsolutePath());

          // ##### Does this need to happen? Follow
          // openStructureFileForSequence() below
          /*
          PDBEntry fileEntry = new AssociatePdbFileWithSeq()
                  .associatePdbWithSeq(structureFile.getAbsolutePath(),
                          DataSourceType.FILE, seq, true, Desktop.instance);
                          */

          // open structure view
          AlignmentPanel ap = af.alignPanel;
          if (headless)
          {
            Cache.setProperty(Preferences.STRUCTURE_DISPLAY,
                    StructureViewer.ViewerType.JMOL.toString());
          }

          String structureFilepath = structureFile.getAbsolutePath();

          // get PAEMATRIX file and label from subvals or Arg.PAEMATRIX
          String paeFilepath = subVals.getWithSubstitutions(argParser, id,
                  "paematrix");
          String paeLabel = subVals.get("paelabel");
          ArgValue paeAv = getArgAssociatedWithStructure(Arg.PAEMATRIX, avm,
                  af, structureFilepath);
          if (paeFilepath == null && paeAv != null)
          {
            SubVals sv = paeAv.getSubVals();
            File paeFile = new File(sv.getContent());

            paeLabel = sv.get("label");
            try
            {
              paeFilepath = paeFile.getCanonicalPath();
            } catch (IOException e)
            {
              paeFilepath = paeFile.getAbsolutePath();
              Console.warn("Problem with the PAE file path: '"
                      + paeFile.getPath() + "'");
            }
          }

          // showing annotations from structure file or not
          boolean ssFromStructure = ArgParser.getFromSubValArgOrPref(avm,
                  Arg.SSANNOTATIONS, subVals, null, "STRUCT_FROM_PDB",
                  true);

          // get TEMPFAC type from subvals or Arg.TEMPFAC in case user Adds
          // reference annotations
          String tftString = subVals.get("tempfac");
          TFType tft = avm.getBoolean(Arg.NOTEMPFAC) ? null
                  : TFType.DEFAULT;
          ArgValue tftAv = getArgAssociatedWithStructure(Arg.TEMPFAC, avm,
                  af, structureFilepath);
          if (tftString == null && tftAv != null)
          {
            tftString = tftAv.getSubVals().getContent();
          }
          if (tftString != null)
          {
            // get kind of temperature factor annotation
            try
            {
              tft = TFType.valueOf(tftString.toUpperCase(Locale.ROOT));
              Console.debug("Obtained Temperature Factor type of '" + tft
                      + "' for structure '" + structureFilepath + "'");
            } catch (IllegalArgumentException e)
            {
              // Just an error message!
              StringBuilder sb = new StringBuilder().append("Cannot set ")
                      .append(Arg.TEMPFAC.argString()).append(" to '")
                      .append(tft)
                      .append("', ignoring.  Valid values are: ");
              Iterator<TFType> it = Arrays.stream(TFType.values())
                      .iterator();
              while (it.hasNext())
              {
                sb.append(it.next().toString().toLowerCase(Locale.ROOT));
                if (it.hasNext())
                  sb.append(", ");
              }
              Console.warn(sb.toString());
            }
          }

          String sViewer = ArgParser.getFromSubValArgOrPref(avm,
                  Arg.STRUCTUREVIEWER, Position.AFTER, av, subVals,
                  "viewer", null, "jmol");
          ViewerType viewerType = null;
          if (!"none".equals(sViewer))
          {
            for (ViewerType v : EnumSet.allOf(ViewerType.class))
            {
              String name = v.name().toLowerCase(Locale.ROOT)
                      .replaceAll(" ", "");
              if (sViewer.equals(name))
              {
                viewerType = v;
                break;
              }
            }
          }

          boolean addTempFac = tft != null
                  || Cache.getDefault("ADD_TEMPFACT_ANN", false);

          // TODO use ssFromStructure
          StructureChooser.openStructureFileForSequence(null, null, ap, seq,
                  false, structureFilepath, tft, paeFilepath, false,
                  ssFromStructure, false, viewerType);
        }
      }
    }

    /*
    boolean doShading = avm.getBoolean(Arg.TEMPFAC_SHADING);
    if (doShading)
    {
      AlignFrame af = afMap.get(id);
      for (AlignmentAnnotation aa : af.alignPanel.getAlignment()
              .findAnnotation(PDBChain.class.getName().toString()))
      {
        AnnotationColourGradient acg = new AnnotationColourGradient(aa,
                af.alignPanel.av.getGlobalColourScheme(), 0);
        acg.setSeqAssociated(true);
        af.changeColour(acg);
        Console.info("Changed colour " + acg.toString());
      }
    }
    */

    return theseArgsWereParsed;
  }

  protected boolean processImages(String id)
  {
    ArgValuesMap avm = argParser.getLinkedArgs(id);
    AlignFrame af = afMap.get(id);

    if (af == null)
    {
      Console.warn("Did not have an alignment window for id=" + id);
      return false;
    }

    if (avm.containsArg(Arg.IMAGE))
    {
      for (ArgValue av : avm.getArgValueList(Arg.IMAGE))
      {
        String val = av.getValue();
        SubVals subVal = av.getSubVals();
        String type = "png"; // default
        String fileName = subVal.getContent();
        File file = new File(fileName);
        String renderer = ArgParser.getValueFromSubValOrArg(avm, av,
                Arg.RENDERER, subVal);
        if (renderer == null)
          renderer = "text";
        type = ArgParser.getValueFromSubValOrArg(avm, av, Arg.TYPE, subVal);
        if (type == null && fileName != null)
        {
          for (String ext : new String[] { "svg", "png", "html", "eps" })
          {
            if (fileName.toLowerCase(Locale.ROOT).endsWith("." + ext))
            {
              type = ext;
            }
          }
        }
        // for moment we disable JSON export
        Cache.setPropsAreReadOnly(true);
        Cache.setProperty("EXPORT_EMBBED_BIOJSON", "false");

        switch (type)
        {

        case "svg":
          Console.debug("Outputting type '" + type + "' to " + fileName);
          af.createSVG(file, renderer);
          break;

        case "png":
          Console.debug("Outputting type '" + type + "' to " + fileName);
          af.createPNG(file);
          break;

        case "html":
          Console.debug("Outputting type '" + type + "' to " + fileName);
          HtmlSvgOutput htmlSVG = new HtmlSvgOutput(af.alignPanel);
          htmlSVG.exportHTML(fileName, renderer);
          break;

        case "biojs":
          try
          {
            BioJsHTMLOutput.refreshVersionInfo(
                    BioJsHTMLOutput.BJS_TEMPLATES_LOCAL_DIRECTORY);
          } catch (URISyntaxException e)
          {
            e.printStackTrace();
          }
          BioJsHTMLOutput bjs = new BioJsHTMLOutput(af.alignPanel);
          bjs.exportHTML(fileName);
          Console.debug("Creating BioJS MSA Viwer HTML file: " + fileName);
          break;

        default:
          Console.warn(Arg.IMAGE.argString() + " type '" + type
                  + "' not known. Ignoring");
          break;
        }
      }
    }
    return true;
  }

  private SequenceI getSpecifiedSequence(AlignFrame af, SubVals subId)
  {
    if (subId == null)
      return null;
    AlignmentI al = af.getCurrentView().getAlignment();
    if (subId.has("seqid"))
    {
      return al.findName(subId.get("seqid"));
    }
    else if (-1 < subId.getIndex()
            && subId.getIndex() < al.getSequences().size())
    {
      return al.getSequenceAt(subId.getIndex());
    }
    return null;
  }

  // returns the first Arg value intended for the structure structFilename
  // (in the given AlignFrame from the ArgValuesMap)
  private ArgValue getArgAssociatedWithStructure(Arg arg, ArgValuesMap avm,
          AlignFrame af, String structFilename)
  {
    if (af != null)
    {
      for (ArgValue av : avm.getArgValueList(arg))
      {
        SubVals subVals = av.getSubVals();
        String structid = subVals.get("structid");
        String structfile = subVals.get("structfile");

        // let's find a structure
        if (structfile == null && structid == null)
        {
          ArgValue likelyStructure = avm.getClosestPreviousArgValueOfArg(av,
                  Arg.STRUCTURE);
          if (likelyStructure != null)
          {
            SubVals sv = likelyStructure.getSubVals();
            if (sv != null && sv.has(ArgValues.ID))
            {
              structid = sv.get(ArgValues.ID);
            }
            else
            {
              structfile = likelyStructure.getValue();
            }
          }
        }

        if (structfile == null && structid != null)
        {
          StructureSelectionManager ssm = StructureSelectionManager
                  .getStructureSelectionManager(Desktop.instance);
          if (ssm != null)
          {
            structfile = ssm.findFileForPDBId(structid);
          }
        }
        if (structfile != null && structfile.equals(structFilename))
        {
          return av;
        }
      }
    }
    return null;
  }
}
